@@ -76,6 +76,7 @@ void nbdtab_set_property(char *property, char *val) {
 	SET_PROP("keyfile", key, val);
 	SET_PROP("cacertfile", cacert, val);
 	SET_PROP("tlshostname", tlshostn, val);
+	SET_PROP("priority", priority, val);
 	SET_PROP("bs", bs, strtol(val, NULL, 10));
 	SET_PROP("timeout", timeout, strtol(val, NULL, 10));
 	SET_PROP("conns", nconn, strtol(val, NULL, 10));
@@ -307,7 +308,7 @@ int opennet(char *name, char* portstr) {
 
 		if(connect(sock, rp->ai_addr, rp->ai_addrlen) != -1)
 			break;		/* success */
-			
+
 		close(sock);
 	}
 
@@ -392,7 +393,15 @@ struct reply {
 
 struct reply* read_reply(int sock) {
 	struct reply *retval = malloc(sizeof(struct reply));
-	readit(sock, retval, sizeof(*retval));
+
+	if (retval == NULL)
+		return NULL;
+
+	if(readit(sock, retval, sizeof(*retval)) < 0) {
+		free(retval);
+		return NULL;
+	}
+
 	retval->magic = ntohll(retval->magic);
 	retval->opt = ntohl(retval->opt);
 	retval->reply_type = ntohl(retval->reply_type);
@@ -401,8 +410,13 @@ struct reply* read_reply(int sock) {
 		fprintf(stderr, "E: received invalid negotiation magic %" PRIu64 " (expected %" PRIu64 ")", retval->magic, rep_magic);
 		exit(EXIT_FAILURE);
 	}
-	if (retval->datasize > 0) {
-		retval = realloc(retval, sizeof(struct reply) + retval->datasize);
+	if (retval->datasize > 0 && retval->datasize < 4096) {
+		struct reply *retval_r = realloc(retval, sizeof(struct reply) + retval->datasize);
+		if (retval_r == NULL) {
+			free(retval);
+			return NULL;
+		}
+		retval = retval_r;
 		readit(sock, &(retval->data), retval->datasize);
 	}
 	return retval;
@@ -482,6 +496,10 @@ void ask_list(int sock) {
 				len -= lenn;
 				len -= sizeof(lenn);
 				if(len > 0) {
+					if(len >= BUF_SIZE) {
+						fprintf(stderr, "\nE: export description read buffer overflow\n");
+						exit(EXIT_FAILURE);
+					}
 					if(read(sock, buf, len) < 0) {
 						fprintf(stderr, "\nE: could not read export description from server\n");
 						exit(EXIT_FAILURE);
@@ -526,7 +544,7 @@ void send_opt_exportname(int sock, u64 *rsize64, uint16_t *flags, bool can_opt_g
 	}
 }
 
-void negotiate(int *sockp, u64 *rsize64, uint16_t *flags, char* name, uint32_t needed_flags, uint32_t client_flags, uint32_t do_opts, char *certfile, char *keyfile, char *cacertfile, char *tlshostname, bool tls, bool can_opt_go) {
+void negotiate(int *sockp, u64 *rsize64, uint16_t *flags, char* name, uint32_t needed_flags, uint32_t client_flags, uint32_t do_opts, char *certfile, char *keyfile, char *cacertfile, char *tlshostname, bool tls, char *priority, bool can_opt_go) {
 	u64 magic;
 	uint16_t tmp;
 	uint16_t global_flags;
@@ -602,6 +620,7 @@ void negotiate(int *sockp, u64 *rsize64, uint16_t *flags, char* name, uint32_t n
 				   certfile,
 				   cacertfile,
 				   tlshostname,
+				   priority,
 				   !cacertfile || !tlshostname, // insecure flag
 #ifdef DODBG
 				   1, // debug
@@ -660,7 +679,7 @@ void negotiate(int *sockp, u64 *rsize64, uint16_t *flags, char* name, uint32_t n
 	}
 
 	struct reply *rep = NULL;
-	
+
 	if(!can_opt_go) {
 		send_opt_exportname(sock, rsize64, flags, can_opt_go, name, global_flags);
 		return;
@@ -671,7 +690,11 @@ void negotiate(int *sockp, u64 *rsize64, uint16_t *flags, char* name, uint32_t n
 	do {
 		if(rep != NULL) free(rep);
 		rep = read_reply(sock);
-		if(rep && (rep->reply_type & NBD_REP_FLAG_ERROR)) {
+		if(rep == NULL) {
+			err("Unable to read reply.");
+			exit(EXIT_FAILURE);
+		}
+		if(rep->reply_type & NBD_REP_FLAG_ERROR) {
 			switch(rep->reply_type) {
 				case NBD_REP_ERR_UNSUP:
 					/* server doesn't support NBD_OPT_GO or NBD_OPT_INFO,
@@ -682,6 +705,7 @@ void negotiate(int *sockp, u64 *rsize64, uint16_t *flags, char* name, uint32_t n
 				case NBD_REP_ERR_POLICY:
 					if(rep->datasize > 0) {
 						char errstr[1024];
+						rep->data[rep->datasize - 1] = '\0';
 						snprintf(errstr, sizeof errstr, "Connection not allowed by server policy. Server said: %s", rep->data);
 						err(errstr);
 					} else {
@@ -692,6 +716,7 @@ void negotiate(int *sockp, u64 *rsize64, uint16_t *flags, char* name, uint32_t n
 				default:
 					if(rep->datasize > 0) {
 						char errstr[1024];
+						rep->data[rep->datasize - 1] = '\0';
 						snprintf(errstr, sizeof errstr, "Unknown error returned by server. Server said: %s", rep->data);
 						err(errstr);
 					} else {
@@ -724,14 +749,14 @@ void negotiate(int *sockp, u64 *rsize64, uint16_t *flags, char* name, uint32_t n
 	free(rep);
 }
 
-bool get_from_config(char* cfgname, char** name_ptr, char** dev_ptr, char** hostn_ptr, int* bs, int* timeout, int* persist, int* swap, int* b_unix, char**port, int* num_conns, char **certfile, char **keyfile, char **cacertfile, char **tlshostname, bool *can_opt_go) {
+bool get_from_config(char* cfgname, char** name_ptr, char** dev_ptr, char** hostn_ptr, int* bs, int* timeout, int* persist, int* swap, int* b_unix, char**port, int* num_conns, char **certfile, char **keyfile, char **cacertfile, char **tlshostname, char **priority, bool *can_opt_go) {
 	bool retval = false;
 	cur_client = calloc(sizeof(CLIENT), 1);
 	cur_client->bs = 512;
 	cur_client->nconn = 1;
 	yyin = fopen(SYSCONFDIR "/nbdtab", "r");
 	yyout = fopen("/dev/null", "w");
-	
+
 	if(!strncmp(cfgname, "/dev/", 5)) {
 		cfgname += 5;
 	}
@@ -748,6 +773,9 @@ bool get_from_config(char* cfgname, char** name_ptr, char** dev_ptr, char** host
 	}
 	*name_ptr = cur_client->name;
 	*dev_ptr = calloc(strlen(cur_client->dev) + 6, 1);
+	if (!*dev_ptr) {
+		goto out;
+	}
 	snprintf(*dev_ptr, strlen(cur_client->dev) + 6, "/dev/%s", cur_client->dev);
 	*hostn_ptr = cur_client->hostn;
 	*bs = cur_client->bs;
@@ -755,11 +783,13 @@ bool get_from_config(char* cfgname, char** name_ptr, char** dev_ptr, char** host
 	*persist = cur_client->persist ? 1 : 0;
 	*swap = cur_client->swap ? 1 : 0;
 	*b_unix = cur_client->b_unix ? 1 : 0;
+	*port = cur_client->port;
 	*num_conns = cur_client->nconn;
 	*certfile = cur_client->cert;
 	*keyfile = cur_client->key;
 	*cacertfile = cur_client->cacert;
 	*tlshostname = cur_client->tlshostn;
+	*priority = cur_client->priority;
 	*can_opt_go = !(cur_client->no_optgo);
 
 	retval = true;
@@ -871,7 +901,7 @@ void usage(char* errmsg, ...) {
 	fprintf(stderr, "Or   : nbd-client -l|--list host\n");
 	fprintf(stderr, "Or   : nbd-client -V|--version\n");
 #if HAVE_GNUTLS && !defined(NOTLS)
-	fprintf(stderr, "All commands that connect to a host also take:\n\t[-F|-certfile certfile] [-K|-keyfile keyfile]\n\t[-A|-cacertfile cacertfile] [-H|-tlshostname hostname] [-x|-enable-tls]\n");
+	fprintf(stderr, "All commands that connect to a host also take:\n\t[-F|-certfile certfile] [-K|-keyfile keyfile]\n\t[-A|-cacertfile cacertfile] [-H|-tlshostname hostname] [-x|-enable-tls]\n\t[-y|-priority gnutls-priority-string]\n");
 #endif
 	fprintf(stderr, "Default value for blocksize is 512\n");
 	fprintf(stderr, "Allowed values for blocksize are 512,1024,2048,4096\n"); /* will be checked in kernel :) */
@@ -896,11 +926,14 @@ void disconnect(char* device) {
 	close(nbd);
 }
 
+static const char *short_opts = "-B:b:c:d:gH:hlnN:PpRSst:uVxy:"
 #if HAVE_NETLINK
-static const char *short_opts = "-A:B:b:c:C:d:gH:hK:LlnN:PpRSst:uVx";
-#else
-static const char *short_opts = "-A:B:b:c:C:d:gH:hK:lnN:PpRSst:uVx";
+	"L"
 #endif
+#if HAVE_GNUTLS
+	"A:C:F:K:"
+#endif
+	;
 
 int main(int argc, char *argv[]) {
 	char* port=NBD_DEFAULT_PORT;
@@ -930,6 +963,7 @@ int main(int argc, char *argv[]) {
 	char *keyfile = NULL;
 	char *cacertfile = NULL;
 	char *tlshostname = NULL;
+	char *priority = NULL;
 	bool tls = false;
 	struct sigaction sa;
 	int num_connections = 1;
@@ -963,7 +997,8 @@ int main(int argc, char *argv[]) {
 		{ "unix", no_argument, NULL, 'u' },
 		{ "version", no_argument, NULL, 'V' },
 		{ "enable-tls", no_argument, NULL, 'x' },
-		{ 0, 0, 0, 0 }, 
+		{ "priority", required_argument, NULL, 'y' },
+		{ 0, 0, 0, 0 },
 	};
 	int i;
 	bool can_opt_go = true;
@@ -1104,11 +1139,15 @@ int main(int argc, char *argv[]) {
                 case 'H':
                         tlshostname=strdup(optarg);
                         break;
+                case 'y':
+			            priority=strdup(optarg);
+			            break;
 #else
                 case 'F':
                 case 'K':
                 case 'H':
                 case 'A':
+		        case 'y':
 			fprintf(stderr, "E: TLS support not compiled in\n");
                         exit(EXIT_FAILURE);
 #endif
@@ -1132,7 +1171,7 @@ int main(int argc, char *argv[]) {
 	if(hostname) {
 		if((!name || !nbddev) && !(opts & NBDC_DO_LIST)) {
 			if(!strncmp(hostname, "nbd", 3) || !strncmp(hostname, "/dev/nbd", 8)) {
-				if(!get_from_config(hostname, &name, &nbddev, &hostname, &blocksize, &timeout, &cont, &swap, &b_unix, &port, &num_connections, &certfile, &keyfile, &cacertfile, &tlshostname, &can_opt_go)) {
+				if(!get_from_config(hostname, &name, &nbddev, &hostname, &blocksize, &timeout, &cont, &swap, &b_unix, &port, &num_connections, &certfile, &keyfile, &cacertfile, &tlshostname, &priority, &can_opt_go)) {
 					usage("no valid configuration for specified device found", hostname);
 					exit(EXIT_FAILURE);
 				}
@@ -1200,7 +1239,7 @@ int main(int argc, char *argv[]) {
 			exit(EXIT_FAILURE);
 
 		if (!preinit)
-			negotiate(&sock, &size64, &flags, name, needed_flags, cflags, opts, certfile, keyfile, cacertfile, tlshostname, tls, can_opt_go);
+			negotiate(&sock, &size64, &flags, name, needed_flags, cflags, opts, certfile, keyfile, cacertfile, tlshostname, tls, priority, can_opt_go);
 		if (force_read_only)
 			flags |= NBD_FLAG_READ_ONLY;
 		if (force_size64)
@@ -1237,7 +1276,7 @@ int main(int argc, char *argv[]) {
 		return 0;
 	}
 	/* Go daemon */
-	
+
 #ifndef NOFORK
 	if(!nofork) {
 		if (daemon(0,0) < 0)
@@ -1318,7 +1357,7 @@ int main(int argc, char *argv[]) {
 					nbd = open(nbddev, O_RDWR);
 					if (nbd < 0)
 						err("Cannot open NBD: %m");
-					negotiate(&sock, &new_size, &new_flags, name, needed_flags, cflags, opts, certfile, keyfile, cacertfile, tlshostname, tls, can_opt_go);
+					negotiate(&sock, &new_size, &new_flags, name, needed_flags, cflags, opts, certfile, keyfile, cacertfile, tlshostname, tls, priority, can_opt_go);
 					if (size64 != new_size) {
 						err("Size of the device changed. Bye");
 					}
