@@ -27,6 +27,7 @@
 #include <string.h>
 #include <sys/time.h>
 #include <sys/types.h>
+#include <sys/wait.h>
 #include <sys/un.h>
 #include <sys/socket.h>
 #include <sys/stat.h>
@@ -60,10 +61,9 @@ static gchar *cacertfile = NULL;
 static gchar *tlshostname = NULL;
 
 typedef enum {
-	CONNECTION_TYPE_NONE,
-	CONNECTION_TYPE_CONNECT,
 	CONNECTION_TYPE_INIT_PASSWD,
 	CONNECTION_TYPE_CLISERV,
+	CONNECTION_TYPE_HANDSHAKE,
 	CONNECTION_TYPE_FULL,
 } CONNECTION_TYPE;
 
@@ -74,7 +74,7 @@ typedef enum {
 
 struct reqcontext {
 	uint64_t seq;
-	char orighandle[8];
+	uint64_t origcookie;
 	struct nbd_request req;
 	struct reqcontext *next;
 	struct reqcontext *prev;
@@ -277,10 +277,9 @@ int writebuffer(int fd, struct chunklist *l)
 	return 0;
 }
 
-#define TEST_WRITE (1<<0)
-#define TEST_FLUSH (1<<1)
-#define TEST_EXPECT_ERROR (1<<2)
-#define TEST_HANDSHAKE (1<<3)
+#define TEST_WRITE              (1<<0)
+#define TEST_FLUSH              (1<<1)
+#define TEST_EXPECT_ERROR       (1<<2)
 
 int timeval_subtract(struct timeval *result, struct timeval *x,
 		     struct timeval *y)
@@ -374,6 +373,7 @@ int setup_connection_common(int sock, char *name, CONNECTION_TYPE ctype,
 		goto end;
 	READ_ALL_ERRCHK(sock, buf, strlen(INIT_PASSWD), err,
 			"Could not read INIT_PASSWD: %s", strerror(errno));
+	buf[strlen(INIT_PASSWD)] = 0;
 	if (strlen(buf) == 0) {
 		snprintf(errstr, errstr_len, "Server closed connection");
 		goto err;
@@ -391,7 +391,7 @@ int setup_connection_common(int sock, char *name, CONNECTION_TYPE ctype,
 		strncpy(errstr, "mymagic does not match", errstr_len);
 		goto err;
 	}
-	if (ctype < CONNECTION_TYPE_FULL)
+	if (ctype < CONNECTION_TYPE_HANDSHAKE)
 		goto end;
 	if (!name) {
 		READ_ALL_ERRCHK(sock, &size, sizeof(size), err,
@@ -420,7 +420,7 @@ int setup_connection_common(int sock, char *name, CONNECTION_TYPE ctype,
 	negotiationflags = htonl(negotiationflags);
 	WRITE_ALL_ERRCHK(sock, &negotiationflags, sizeof(negotiationflags), err,
 			 "Could not write reserved field: %s", strerror(errno));
-	if (testflags & TEST_HANDSHAKE) {
+	if (ctype < CONNECTION_TYPE_FULL) {
 		/* Server must support newstyle for this test */
 		if (!(handshakeflags & NBD_FLAG_FIXED_NEWSTYLE)) {
 			strncpy(errstr, "server does not support handshake", errstr_len);
@@ -486,6 +486,7 @@ int setup_connection_common(int sock, char *name, CONNECTION_TYPE ctype,
 				   certfile,
 				   cacertfile,
 				   tlshostname,
+				   NULL, // use default priority string
 				   !cacertfile || !tlshostname, // insecure flag
 #ifdef DODBG
 				   1, // debug
@@ -539,7 +540,7 @@ int setup_connection_common(int sock, char *name, CONNECTION_TYPE ctype,
 #endif
 	if(testflags & TEST_EXPECT_ERROR) {
 		struct sigaction act;
-		memset(&act, '0', sizeof act);
+		memset(&act, 0, sizeof act);
 		act.sa_handler = SIG_IGN;
 		sigaction(SIGPIPE, &act, NULL);
 	}
@@ -574,16 +575,12 @@ end:
 	return sock;
 }
 
-int setup_unix_connection(gchar * unixsock, gchar * name, CONNECTION_TYPE ctype,
-			  int *serverflags, int testflags)
+int setup_unix_connection(gchar * unixsock)
 {
 	struct sockaddr_un addr;
 	int sock;
 
 	sock = 0;
-	if (ctype < CONNECTION_TYPE_CONNECT) {
-		goto end;
-	}
 	if ((sock = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
 		strncpy(errstr, strerror(errno), errstr_len);
 		goto err;
@@ -598,7 +595,6 @@ int setup_unix_connection(gchar * unixsock, gchar * name, CONNECTION_TYPE ctype,
 		strncpy(errstr, strerror(errno), errstr_len);
 		goto err_open;
 	}
-	sock = setup_connection_common(sock, name, ctype, serverflags, testflags);
 	goto end;
 err_open:
 	close(sock);
@@ -608,16 +604,13 @@ end:
 	return sock;
 }
 
-int setup_inet_connection(gchar * hostname, int port, gchar * name,
-			  CONNECTION_TYPE ctype, int *serverflags, int testflags)
+int setup_inet_connection(gchar * hostname, int port)
 {
 	int sock;
 	struct hostent *host;
 	struct sockaddr_in addr;
 
 	sock = 0;
-	if (ctype < CONNECTION_TYPE_CONNECT)
-		goto end;
 	if ((sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
 		strncpy(errstr, strerror(errno), errstr_len);
 		goto err;
@@ -630,11 +623,11 @@ int setup_inet_connection(gchar * hostname, int port, gchar * name,
 	addr.sin_family = AF_INET;
 	addr.sin_port = htons(port);
 	addr.sin_addr.s_addr = *((int *)host->h_addr);
+	memset(&addr.sin_zero, 0, sizeof(addr.sin_zero));
 	if ((connect(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0)) {
 		strncpy(errstr, strerror(errno), errstr_len);
 		goto err_open;
 	}
-	sock = setup_connection_common(sock, name, ctype, serverflags, testflags);
 	goto end;
 err_open:
 	close(sock);
@@ -644,19 +637,39 @@ end:
 	return sock;
 }
 
-int setup_connection(gchar * hostname, gchar * unixsock, int port, gchar * name,
-		     CONNECTION_TYPE ctype, int *serverflags, int testflags)
+int setup_inetd_connection(gchar **argv)
 {
-	if (hostname != NULL) {
-		return setup_inet_connection(hostname, port, name, ctype,
-					     serverflags, testflags);
-	} else if (unixsock != NULL) {
-		return setup_unix_connection(unixsock, name, ctype,
-					     serverflags, testflags);
-	} else {
-		g_error("need a hostname or a unix domain socket!");
+	int sv[2], status;
+	pid_t child;
+
+	if (socketpair(AF_UNIX, SOCK_STREAM, 0, sv) == -1) {
+		strncpy(errstr, strerror(errno), errstr_len);
 		return -1;
 	}
+
+	child = vfork();
+	if (child == 0) {
+		dup2(sv[0], 0);
+		close(sv[0]);
+		close(sv[1]);
+		execvp(argv[0], argv);
+		perror("execvp");
+		_exit(-1);
+	} else if (child == -1) {
+		close(sv[0]);
+		close(sv[1]);
+		strncpy(errstr, strerror(errno), errstr_len);
+		return -1;
+	}
+
+	close(sv[0]);
+	if (waitpid(child, &status, WNOHANG)) {
+		close(sv[1]);
+		return -1;
+	}
+
+	setmysockopt(sv[1]);
+	return sv[1];
 }
 
 int close_connection(int sock, CLOSE_TYPE type)
@@ -668,7 +681,7 @@ int close_connection(int sock, CLOSE_TYPE type)
 	case CONNECTION_CLOSE_PROPERLY:
 		req.magic = htonl(NBD_REQUEST_MAGIC);
 		req.type = htonl(NBD_CMD_DISC);
-		memcpy(&(req.handle), &(counter), sizeof(counter));
+		req.cookie = htonll(counter);
 		counter++;
 		req.from = 0;
 		req.len = 0;
@@ -678,6 +691,7 @@ int close_connection(int sock, CLOSE_TYPE type)
 				 strerror(errno));
 			return -1;
 		}
+		/* falls through */
 	case CONNECTION_CLOSE_FAST:
 		if (close(sock) < 0) {
 			snprintf(errstr, errstr_len,
@@ -692,7 +706,7 @@ int close_connection(int sock, CLOSE_TYPE type)
 	return 0;
 }
 
-int read_packet_check_header(int sock, size_t datasize, long long int curhandle)
+int read_packet_check_header(int sock, size_t datasize, long long int curcookie)
 {
 	struct nbd_reply rep;
 	int retval = 0;
@@ -704,11 +718,11 @@ int read_packet_check_header(int sock, size_t datasize, long long int curhandle)
 	rep.error = ntohl(rep.error);
 	if (rep.magic != NBD_REPLY_MAGIC) {
 		snprintf(errstr, errstr_len,
-			 "Received package with incorrect reply_magic. Index of sent packages is %lld (0x%llX), received handle is %lld (0x%llX). Received magic 0x%lX, expected 0x%lX",
-			 (long long int)curhandle,
-			 (long long unsigned int)curhandle,
-			 (long long int)*((u64 *) rep.handle),
-			 (long long unsigned int)*((u64 *) rep.handle),
+			 "Received package with incorrect reply_magic. Index of sent packages is %lld (0x%llX), received cookie is %lld (0x%llX). Received magic 0x%lX, expected 0x%lX",
+			 (long long int)curcookie,
+			 (long long unsigned int)curcookie,
+			 (long long int)rep.cookie,
+			 (long long unsigned int)rep.cookie,
 			 (long unsigned int)rep.magic,
 			 (long unsigned int)NBD_REPLY_MAGIC);
 		retval = -1;
@@ -716,10 +730,10 @@ int read_packet_check_header(int sock, size_t datasize, long long int curhandle)
 	}
 	if (rep.error) {
 		snprintf(errstr, errstr_len,
-			 "Received error from server: %ld (0x%lX). Handle is %lld (0x%llX).",
+			 "Received error from server: %ld (0x%lX). Cookie is %lld (0x%llX).",
 			 (long int)rep.error, (long unsigned int)rep.error,
-			 (long long int)(*((u64 *) rep.handle)),
-			 (long long unsigned int)*((u64 *) rep.handle));
+			 (long long int)rep.cookie,
+			 (long long unsigned int)rep.cookie);
 		retval = -2;
 		goto end;
 	}
@@ -731,8 +745,7 @@ end:
 	return retval;
 }
 
-int oversize_test(gchar * hostname, gchar * unixsock, int port, char *name,
-		  int sock, char sock_is_open, char close_sock, int testflags)
+int oversize_test(char *name, int sock, char close_sock, int testflags)
 {
 	int retval = 0;
 	struct nbd_request req;
@@ -744,20 +757,18 @@ int oversize_test(gchar * hostname, gchar * unixsock, int port, char *name,
 	bool got_err;
 
 	/* This should work */
-	if (!sock_is_open) {
-		if ((sock =
-		     setup_connection(hostname, unixsock, port, name,
-				      CONNECTION_TYPE_FULL,
-				      &serverflags, testflags)) < 0) {
-			g_warning("Could not open socket: %s", errstr);
-			retval = -1;
-			goto err;
-		}
+	if ((sock =
+		 setup_connection_common(sock, name,
+				  CONNECTION_TYPE_FULL,
+				  &serverflags, testflags)) < 0) {
+		g_warning("Could not open socket: %s", errstr);
+		retval = -1;
+		goto err;
 	}
 	req.magic = htonl(NBD_REQUEST_MAGIC);
 	req.type = htonl(NBD_CMD_READ);
 	req.len = htonl(1024 * 1024);
-	memcpy(&(req.handle), &i, sizeof(i));
+	req.cookie = htonll(i);
 	req.from = htonll(i);
 	WRITE_ALL_ERR_RT(sock, &req, sizeof(req), err, -1,
 			 "Could not write request: %s", strerror(errno));
@@ -816,8 +827,7 @@ err:
 	return retval;
 }
 
-int handshake_test(gchar * hostname, gchar * unixsock, int port, char *name,
-		   int sock, char sock_is_open, char close_sock, int testflags)
+int handshake_test(char *name, int sock, char close_sock, int testflags)
 {
 	int retval = -1;
 	int serverflags = 0;
@@ -825,14 +835,12 @@ int handshake_test(gchar * hostname, gchar * unixsock, int port, char *name,
 	uint32_t tmp32 = 0;
 
 	/* This should work */
-	if (!sock_is_open) {
-		if ((sock =
-		     setup_connection(hostname, unixsock, port, name,
-				      CONNECTION_TYPE_FULL,
-				      &serverflags, testflags)) < 0) {
-			g_warning("Could not open socket: %s", errstr);
-			goto err;
-		}
+	if ((sock =
+		setup_connection_common(sock, name,
+				  CONNECTION_TYPE_HANDSHAKE,
+				  &serverflags, testflags)) < 0) {
+		g_warning("Could not open socket: %s", errstr);
+		goto err;
 	}
 
 	/* Intentionally throw an unknown option at the server */
@@ -899,8 +907,7 @@ err:
 	return retval;
 }
 
-int throughput_test(gchar * hostname, gchar * unixsock, int port, char *name,
-		    int sock, char sock_is_open, char close_sock, int testflags)
+int throughput_test(char *name, int sock, char close_sock, int testflags)
 {
 	long long int i;
 	char writebuf[1024];
@@ -924,20 +931,18 @@ int throughput_test(gchar * hostname, gchar * unixsock, int port, char *name,
 
 	memset(writebuf, 'X', 1024);
 	size = 0;
-	if (!sock_is_open) {
-		if ((sock =
-		     setup_connection(hostname, unixsock, port, name,
-				      CONNECTION_TYPE_FULL,
-				      &serverflags, testflags)) < 0) {
-			g_warning("Could not open socket: %s", errstr);
-			if(testflags & TEST_EXPECT_ERROR) {
-				g_message("Test failed, as expected");
-				retval = 0;
-			} else {
-				retval = -1;
-			}
-			goto err;
+	if ((sock =
+		 setup_connection_common(sock, name,
+				  CONNECTION_TYPE_FULL,
+				  &serverflags, testflags)) < 0) {
+		g_warning("Could not open socket: %s", errstr);
+		if(testflags & TEST_EXPECT_ERROR) {
+			g_message("Test failed, as expected");
+			retval = 0;
+		} else {
+			retval = -1;
 		}
+		goto err;
 	}
 	if ((testflags & TEST_FLUSH)
 	    && ((serverflags & (NBD_FLAG_SEND_FLUSH | NBD_FLAG_SEND_FUA))
@@ -967,7 +972,7 @@ int throughput_test(gchar * hostname, gchar * unixsock, int port, char *name,
 			if (sendfua)
 				req.type =
 				    htonl(NBD_CMD_WRITE | NBD_CMD_FLAG_FUA);
-			memcpy(&(req.handle), &i, sizeof(i));
+			req.cookie = htonll(i);
 			req.from = htonll(i);
 			if (write_all(sock, &req, sizeof(req)) < 0) {
 				retval = -1;
@@ -983,7 +988,7 @@ int throughput_test(gchar * hostname, gchar * unixsock, int port, char *name,
 			if (sendflush) {
 				long long int j = i ^ (1LL << 63);
 				req.type = htonl(NBD_CMD_FLUSH);
-				memcpy(&(req.handle), &j, sizeof(j));
+				req.cookie = htonll(j);
 				req.from = 0;
 				req.len = 0;
 				if (write_all(sock, &req, sizeof(req)) < 0) {
@@ -1100,8 +1105,8 @@ err:
 
 /*
  * fill 512 byte buffer 'buf' with a hashed selection of interesting data based
- * only on handle and blknum. The first word is blknum, and the second handle, for ease
- * of understanding. Things with handle 0 are blank.
+ * only on cookie and blknum. The first word is blknum, and the second cookie, for ease
+ * of understanding. Things with cookie 0 are blank.
  */
 static inline void makebuf(char *buf, uint64_t seq, uint64_t blknum)
 {
@@ -1157,21 +1162,20 @@ static inline void dumpcommand(char *text, uint32_t command)
 #endif
 }
 
-/* return an unused handle */
-uint64_t getrandomhandle(GHashTable * phash)
+/* return an unused cookie */
+uint64_t getrandomcookie(GHashTable * phash)
 {
-	uint64_t handle = 0;
+	uint64_t cookie = 0;
 	int i;
 	do {
 		/* RAND_MAX may be as low as 2^15 */
 		for (i = 1; i <= 5; i++)
-			handle ^= random() ^ (handle << 15);
-	} while (g_hash_table_lookup(phash, &handle));
-	return handle;
+			cookie ^= random() ^ (cookie << 15);
+	} while (g_hash_table_lookup(phash, &cookie));
+	return cookie;
 }
 
-int integrity_test(gchar * hostname, gchar * unixsock, int port, char *name,
-		   int sock, char sock_is_open, char close_sock, int testflags)
+int integrity_test(char *name, int sock, char close_sock, int testflags)
 {
 	struct nbd_reply rep;
 	fd_set rset;
@@ -1200,17 +1204,15 @@ int integrity_test(gchar * hostname, gchar * unixsock, int port, char *name,
 	struct rclist inflight = { NULL, NULL, 0 };
 	struct chunklist txbuf = { NULL, NULL, 0 };
 
-	GHashTable *handlehash = g_hash_table_new(g_int64_hash, g_int64_equal);
+	GHashTable *cookiehash = g_hash_table_new(g_int64_hash, g_int64_equal);
 
 	size = 0;
-	if (!sock_is_open) {
-		if ((sock =
-		     setup_connection(hostname, unixsock, port, name,
-				      CONNECTION_TYPE_FULL,
-				      &serverflags, testflags)) < 0) {
-			g_warning("Could not open socket: %s", errstr);
-			goto err;
-		}
+	if ((sock =
+		 setup_connection_common(sock, name,
+				  CONNECTION_TYPE_FULL,
+				  &serverflags, testflags)) < 0) {
+		g_warning("Could not open socket: %s", errstr);
+		goto err;
 	}
 
 	if ((serverflags & (NBD_FLAG_SEND_FLUSH | NBD_FLAG_SEND_FUA))
@@ -1333,7 +1335,7 @@ int integrity_test(gchar * hostname, gchar * unixsock, int port, char *name,
 						"Could not read transaction log: %s",
 						strerror(errno));
 				prc->req.magic = htonl(NBD_REQUEST_MAGIC);
-				memcpy(prc->orighandle, prc->req.handle, 8);
+				prc->origcookie = htonll(prc->req.cookie);
 				prc->seq = seq++;
 				if ((ntohl(prc->req.type) &
 				     NBD_CMD_MASK_COMMAND) == NBD_CMD_DISC) {
@@ -1437,10 +1439,9 @@ int integrity_test(gchar * hostname, gchar * unixsock, int port, char *name,
 				rclist_addtail(&inflight, prc);
 
 				dumpcommand("Sending command", prc->req.type);
-				/* we rewrite the handle as they otherwise may not be unique */
-				*((uint64_t *) (prc->req.handle)) =
-				    getrandomhandle(handlehash);
-				g_hash_table_insert(handlehash, prc->req.handle,
+				/* we rewrite the cookie as they otherwise may not be unique */
+				prc->req.cookie = getrandomcookie(cookiehash);
+				g_hash_table_insert(cookiehash, &prc->req.cookie,
 						    prc);
 				addbuffer(&txbuf, &(prc->req),
 					  sizeof(struct nbd_request));
@@ -1536,21 +1537,19 @@ skipdequeue:
 				goto err_open;
 			}
 
-			uint64_t handle;
-			memcpy(&handle, rep.handle, 8);
-			prc = g_hash_table_lookup(handlehash, &handle);
+			uint64_t cookie;
+			cookie = rep.cookie;
+			prc = g_hash_table_lookup(cookiehash, &cookie);
 			if (!prc) {
 				snprintf(errstr, errstr_len,
-					 "Unrecognised handle in reply: 0x%llX",
-					 *(long long unsigned int *)(rep.
-								     handle));
+					 "Unrecognised cookie in reply: 0x%llX",
+					 (long long unsigned int)rep.cookie);
 				goto err_open;
 			}
-			if (!g_hash_table_remove(handlehash, &handle)) {
+			if (!g_hash_table_remove(cookiehash, &cookie)) {
 				snprintf(errstr, errstr_len,
-					 "Could not remove handle from hash: 0x%llX",
-					 *(long long unsigned int *)(rep.
-								     handle));
+					 "Could not remove cookie from hash: 0x%llX",
+					 (long long unsigned int)rep.cookie);
 				goto err_open;
 			}
 
@@ -1695,7 +1694,7 @@ err:
 	if (*errstr)
 		g_warning("%s", errstr);
 
-	g_hash_table_destroy(handlehash);
+	g_hash_table_destroy(cookiehash);
 
 	return retval;
 }
@@ -1720,14 +1719,14 @@ void handle_nonopt(char *opt, gchar ** hostname, long int *p)
 	}
 }
 
-typedef int (*testfunc) (gchar *, gchar *, int, char *, int, char, char, int);
+typedef int (*testfunc) (char *, int, char, int);
 
 int main(int argc, char **argv)
 {
 	gchar *hostname = NULL, *unixsock = NULL;
-	long int p = 0;
+	long int p = 10809;
 	char *name = NULL;
-	int sock = 0;
+	int sock = -1;
 	int c;
 	int testflags = 0;
 	testfunc test = throughput_test;
@@ -1752,16 +1751,13 @@ int main(int argc, char **argv)
 		exit(EXIT_FAILURE);
 	}
 	logging(MY_NAME);
-	while ((c = getopt(argc, argv, "FN:t:owfilu:hC:K:A:H:")) >= 0) {
+	while ((c = getopt(argc, argv, "FN:t:owfilu:hC:K:A:H:I")) >= 0) {
 		switch (c) {
 		case 1:
 			handle_nonopt(optarg, &hostname, &p);
 			break;
 		case 'N':
 			name = g_strdup(optarg);
-			if (!p) {
-				p = 10809;
-			}
 			break;
 		case 'F':
 			testflags |= TEST_EXPECT_ERROR;
@@ -1781,6 +1777,14 @@ int main(int argc, char **argv)
 		case 'f':
 			testflags |= TEST_FLUSH;
 			break;
+		case 'I':
+#ifndef ISSERVER
+			err_nonfatal("inetd mode not supported without syslog support");
+			return 77;
+#else
+			p = -1;
+			break;
+#endif
 		case 'i':
 			test = integrity_test;
 			break;
@@ -1789,7 +1793,6 @@ int main(int argc, char **argv)
 			break;
 		case 'h':
 			test = handshake_test;
-			testflags |= TEST_HANDSHAKE;
 			break;
 #if HAVE_GNUTLS
 		case 'C':
@@ -1816,8 +1819,10 @@ int main(int argc, char **argv)
 		}
 	}
 
-	while (optind < argc) {
-		handle_nonopt(argv[optind++], &hostname, &p);
+	if (p != -1) {
+		while (optind < argc) {
+			handle_nonopt(argv[optind++], &hostname, &p);
+		}
 	}
 
 	if (keyfile && !certfile)
@@ -1826,7 +1831,23 @@ int main(int argc, char **argv)
 	if (!tlshostname && hostname)
 		tlshostname = g_strdup(hostname);
 
-	if (test(hostname, unixsock, (int)p, name, sock, FALSE, TRUE, testflags)
+	if (hostname != NULL) {
+		sock = setup_inet_connection(hostname, p);
+	} else if (unixsock != NULL) {
+		sock = setup_unix_connection(unixsock);
+	} else if (p == -1) {
+		sock = setup_inetd_connection(argv + optind);
+	} else {
+		g_error("need a hostname, a unix domain socket or inetd-mode command line!");
+		return -1;
+	}
+
+	if (sock == -1) {
+		g_warning("Could not establish a connection: %s", errstr);
+		exit(EXIT_FAILURE);
+	}
+
+	if (test(name, sock, TRUE, testflags)
 	    < 0) {
 		g_warning("Could not run test: %s", errstr);
 		exit(EXIT_FAILURE);
